/*
 * Task handling.
 *
 * Copyright (C) 2015, 2016 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "cpu.h"
#include "paging.h"
#include "tasks.h"

/* Task tables and data. */

static u32 stack_pointers[max_tasks];
static u32 registers[max_tasks][32];

u8 current_task;
u32 *current_stack_pointer;
u32 *current_registers;

/* Address locations and paging configuration. */

const u32 stack_start = 0x00080000;
const u32 stack_size = 0x00002000;
const u32 pagesize = 4 * 1024;

/* A reference to the unrelocated symbol table location. */

extern u32 _got_copy_start;

/* Task management functions. */

void init_tasks()
{
	current_task = 0;
	init_task();
}

void init_task()
{
	current_stack_pointer = &stack_pointers[current_task];
	current_registers = registers[current_task];
}

void start_task(unsigned short task, void (*function)(), u32 args[], u8 nargs)
{
	u32 virtual, physical;

	/*
	Each task employs a stack at a multiple of the given start address in
	physical memory, but at the same address in virtual memory.
	*/

	virtual = stack_start;
	physical = stack_start - stack_size * task;

	init_page_table(page_table_start, virtual - pagesize * 2, physical - pagesize * 2, pagesize, 0x1e, task);

	/*
	Subtract from the stack pointer to prevent the called function from
	reaching into unmapped memory.
	*/

	stack_pointers[task] = virtual - 12;

	/*
	Set the registers for the new task, initialising the global pointer and
	return address.
	*/

	init_registers(registers[task], _got_copy_start, function, args, nargs);
}

void start_task_now()
{
	invoke_task(current_registers);
}

void switch_task()
{
	/* Switch the current task. */

	current_task++;
	if (current_task == max_tasks) current_task = 0;
	init_task();
}
